unit ClaKey;

interface

 uses Dialogs, Classes, ClaStru, SysUtils, DB;

 type
   TClarionKey = class(TObject)
    private
     FFile: TFileStream;
     FValidHeader: Boolean;
     FNodeNum: LongInt;
     FPos: Integer;
     FCurRec: LongInt;
     FHeader: TIndexHeader;
     FNode: TKeyNode;

     procedure ReadNode(Num: LongInt; var Buf);
     procedure WriteNode(Num: LongInt; var Buf);
     procedure ReadHeader;
     function GetKey(i: integer): PChar;
     function GetPtr(i: integer): LongInt;

    public
     constructor Create(const AName: string; AMode: Word);
     destructor Destroy; override;
     procedure UpdateHeader;
     function First: LongInt;
     function Last: LongInt;
     function Next: LongInt;
     function Prev: LongInt;
     function Locate(const key: PChar): LongInt;
     function FindRec(Rec: LongInt; Key:PChar): Boolean;

     property CurRec: LongInt read FCurRec;
   end;

implementation

uses Clarion;

function CmpKey(const Key1, Key2: PChar; Len: Integer): Integer; assembler;
asm
   PUSH    EDI
   PUSH    ESI
   MOV     EDI,Key2
   MOV     ESI,Key1
   MOV     ECX,Len
   REPE    CMPSB
   XOR     EAX,EAX
   MOV     AL,[ESI-1]
   XOR     EDX,EDX
   MOV     DL,[EDI-1]
   SUB     EAX,EDX
   POP     ESI
   POP     EDI
end;

constructor TClarionKey.Create(const AName: string; AMode: Word);
 begin
   try
     FFile:=TFileStream.Create(AName, AMode);
     ReadHeader
   except
     FFile:=nil
   end
 end;

destructor TClarionKey.Destroy;
 begin
   FFile.Free
 end;

procedure TClarionKey.ReadNode(Num: LongInt; var Buf);
 begin
   FNodeNum:=Num;
   try
     FFile.Seek(Num*512, soFromBeginning);
     FFile.ReadBuffer(Buf, 512)
   except
     DatabaseError(erInvalidKey)
   end
 end;

procedure TClarionKey.WriteNode(Num: LongInt; var Buf);
 begin
   try
     FFile.Seek(Num*512, soFromBeginning);
     FFile.WriteBuffer(Buf, 512)
   except
     DatabaseError(erInvalidKey)
   end
 end;

procedure TClarionKey.UpdateHeader;
 begin
   FValidHeader:=False
 end;

procedure TClarionKey.ReadHeader;
 begin
   ReadNode(0, FHeader);
   FValidHeader:=True
 end;

function TClarionKey.GetKey(i: integer): PChar;
 begin
   Result:=@FNode.Keys[i*FHeader.KeyLen+4]
 end;

function TClarionKey.GetPtr(i: integer): LongInt;
 begin
   Result:=PLongInt(@FNode.Keys[i*FHeader.KeyLen])^
 end;

function TClarionKey.First: LongInt;
 var L: Integer;
     N: LongInt;
 begin
   if not FValidHeader then ReadHeader;
   if FHeader.NumKEnt<=0 then begin Result:=0; exit end;
   L:=FHeader.NumLvls;
   N:=FHeader.Root;
   repeat
     Dec(L);
     ReadNode(N, FNode);
     N:=GetPtr(0)
   until L=0;
   FPos:=0;
   FCurRec:=N;
   Result:=N
 end;

function TClarionKey.Last: LongInt;
 begin
   if not FValidHeader then ReadHeader;
   if FHeader.NumKEnt<=0 then Result:=0 else
    begin
      ReadNode(FHeader.LastNod, FNode);
      FPos:=FNode.KeyCnt-1;
      Result:=GetPtr(FPos);
      FCurRec:=Result
    end
 end;

function TClarionKey.Next: LongInt;
 begin
   Result:=0;
   if FNodeNum<=0 then exit;
   if FPos<FNode.KeyCnt-1 then Inc(FPos) else
    begin
      if FNode.FLink=0 then exit;
      ReadNode(FNode.FLink, FNode);
      FPos:=0
    end;
   Result:=GetPtr(FPos);
   FCurRec:=Result
 end;

function TClarionKey.Prev: LongInt;
 begin
   Result:=0;
   if FNodeNum<=0 then exit;
   if FPos>0 then Dec(FPos) else
    begin
      if FNode.BLink=0 then exit;
      ReadNode(FNode.BLink, FNode);
      FPos:=FNode.KeyCnt-1
    end;
   Result:=GetPtr(FPos);
   FCurRec:=Result
 end;

function TClarionKey.Locate(const key: PChar): LongInt;
 var L, i, r: Integer;
     N: LongInt;
 begin
   if not FValidHeader then ReadHeader;
   if FHeader.NumKEnt<=0 then begin Result:=0; exit end;
   L:=FHeader.NumLvls;
   N:=FHeader.Root;
   repeat
     ReadNode(N, FNode);
     for i:=0 to FNode.KeyCnt-1 do
      begin
        r:=CmpKey(Key, GetKey(i), FHeader.KeyLen-4);
        if r<=0 then begin N:=GetPtr(i); break end
      end;
     if r>0 then begin Last; Result:=0; Exit end;
     Dec(L)
   until L=0;
   FPos:=i;
   FCurRec:=N;
   if r<>0 then Result:=0 else Result:=N
 end;

function TClarionKey.FindRec(Rec: LongInt; Key:PChar): Boolean;
 begin
   Result:=Locate(Key)<>0;
   if Result then
    repeat
      Result:=FCurRec=Rec;
      if Result then Break;
      if Next=0 then Break
    until CmpKey(Key, GetKey(FPos), FHeader.KeyLen-4)<>0
 end;

end.

