{ TClarionTable }

unit Clarion;

INTERFACE

{$I VERSIONS.INC}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Db,
  Variants, DBConsts, bdeConst,
  {$IFDEF D6PLUS}
  FmtBcd,
  {$ENDIF}
  ClaStru, ClaKey;

type
  PRecInfo = ^TRecInfo;
  TRecInfo = packed record
    Bookmark: longint;
    BookmarkFlag: TBookmarkFlag;
    OldMemoPtr: LongInt;
    MemoBuf: array of Char;
    Updated: boolean
  end;

(* TCDSKey *)

  PCDSKey = ^TCDSKey;
  TCDSKey = record
    FKeyRecord: TKeyRecord;
    FKeyItems: array of TKeyItem;
    Key: TClarionKey
  end;

(* TCDSArray *)

  PCDSArray = ^TCDSArray;
  TCDSArray = record
    FArrayRecord: TArrayRecord;
    FArrayItems: array of TArrayItem;
  end;

  TCharSet = (csOEMCharSet, csANSICharSet);
  TIndexName = type string;

(* TClarion *)

  TClarionTable = class(TDataSet)
  private
    FHeader: THeader;
    FMemoHeader: TMemoHeader;
    FFile: TFileStream;
    FMemo: TFileStream;
    FTableName: TFileName;
    FPassword: String;
    FID: Word;
    FBookmarkOfs: LongInt;
    FExclusive: Boolean;
    FReadOnly: Boolean;
    FOpenMode: Byte;
    FShareMode: Byte;
    FShowDeleted: Boolean;
    FFileUpdated: Boolean;
    FLastDate: LongInt;
    FLastTime: LongInt;
    FCDSFields: array of TFieldRecord;
    FCDSKeys: array of TCDSKey;
    FCDSPictures: array of TPictureRecord;
    FCDSArrays: array of TCDSArray;
    FCurRec: LongInt;
    FCharSet: TCharSet;
    FAutoDate: Boolean;
    FAutoTime: Boolean;
    FStoreDefs: Boolean;
    FDatabaseName: string;
    FIndexDefs: TIndexDefs;
    FFieldsIndex: Boolean;
    FIndexName: TIndexName;
    FFilterOn: Boolean;
    FFilterBuffer: PChar;
    FActiveKey: PCDSKey;
    FKeyBuffer: PChar;
    FSaveKeyBuffer: PChar;

    procedure SetDatabaseName(const Value: string);
    procedure SetTableName(const Value: TFileName);
    procedure SetPassword(const APwd : String);
    procedure SetExclusive(Value : Boolean);
    procedure InternalReadHeader;
    procedure InternalWriteHeader;
    procedure DecodeHeader;
    function CheckHeader: Boolean;
    function CalcCheckSum: Word;
    procedure InternalOpenFile;
    procedure InternalCloseFile;
    procedure SetReadOnly(Value : Boolean);
    procedure SetCharSet(CharSet: TCharSet);
    procedure ReadRecord(RecNum: longint; var Buffer; Len: Word=0);
    procedure WriteRecord(RecNum: longint; const Buffer; Len: Word=0);
    procedure RecordPostAppend(RecBuffer: PDataHeader; EditMode: boolean);
    procedure ReadMemoChunk(RPtr: Integer; var Chunk; Size: Word = 256);
    procedure WriteMemoChunk(RPtr: Integer; var Chunk; Size: Word = 256);
    procedure DeleteMemo(Ptr: LongInt);
    procedure PrepareMemoBuffer;
    procedure FDT2FC(FD: TFieldDef; var FR: TFieldRecord);
    function FieldDefsStored: Boolean;
    function IndexDefsStored: Boolean;
    procedure SetIndexDefs(Value: TIndexDefs);
    procedure SetIndexFieldNames(const Value: string);
    procedure SetIndexName(const Value: string);
    function GetIndexFieldNames: string;
    function GetIndexName: string;
    procedure SetIndex(const Value: string; FieldsIndex: Boolean);
    function FilterRecord(Buf: PChar): boolean;
    function EvaluateFilter: boolean;
    function CurrentBuffer: PChar;
    procedure SelectIndex;
    function CreateKey(Buf: PChar): string;
    procedure ClearRecord(Buffer: PChar);
    procedure SetKeyBuffer(Clear: Boolean);
    procedure SetKeyFields(const Values: array of const);
    procedure PostKeyBuffer(Commit: Boolean);

  protected
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    function GetRecordSize: Word; override;
    procedure InternalClose; override;
    procedure InternalOpen; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalCancel; override;
//    procedure InternalEdit; override;
//    procedure InternalInsert; override;
    procedure InternalRefresh; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalPost; override;
    procedure InternalHandleException; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function IsCursorOpen: Boolean; override;
    function GetRecordCount: LongInt; override;
    function GetRecNo: LongInt; override;
    procedure SetRecNo(Value : LongInt); override;
    function GetCanModify: Boolean; override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    procedure DefChanged(Sender: TObject); override;
    function LocateRecord(const KeyFields: string; const KeyValues: Variant;
                        Options: TLocateOptions; SyncCursor: Boolean): Boolean;
    procedure SetFiltered(Value: Boolean); override;
    procedure SetFilterOptions(Value: TFilterOptions); override;
    procedure SetFilterText(const Value: string); override;
    function FindRecord(Restart, GoForward: Boolean): Boolean; override;
    procedure SetFilterData(const Text: string; Options: TFilterOptions);
    procedure ClearCalcFields(Buffer: PChar); override;

  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    function IsTableLocked : Boolean;
    function IsTableOwned : Boolean;
    function IsTableEncrypted : Boolean;
    function IsTableHaveMemo : Boolean;
    function IsTableCompressed : Boolean;
    function IsTableReclaim : Boolean;
    function IsTableReadonly : Boolean;
    function IsTableCreated : Boolean;
    function IsRecordNew : Boolean;
    function IsRecordOld : Boolean;
    function IsRecordRevised : Boolean;
    function IsRecordDeleted : Boolean;
    function IsRecordHeld : Boolean;
    function Translate(Src, Dest: PChar; ToOem: Boolean): Integer; override;
    function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
    function BookmarkValid(Bookmark: TBookmark): Boolean; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean; override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    procedure CreateTable; overload;
    procedure CreateTable(const Prefix: Char3); overload;
    procedure EmptyTable;
    function GetIsIndexField(Field: TField): Boolean; override;
    procedure UpdateIndexDefs; override;
    function Locate(const KeyFields: string; const KeyValues: Variant;
                    Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant;
                     const ResultFields: string): Variant; override;
    function IsSequenced: Boolean; override;
    procedure SetKey;
    procedure EditKey;
    function GotoKey: Boolean;
    procedure GotoNearest;
    function FindKey(const KeyValues: array of const): Boolean;
    procedure FindNearest(const KeyValues: array of const);
    procedure Post; override;
    procedure Cancel; override;

  published
    property DatabaseName: string read FDatabaseName write SetDatabaseName;
    property TableName: TFileName read FTableName write SetTableName;
    property ShowDeleted: Boolean read FShowDeleted write FShowDeleted default false;
    property Password: String read FPassword write SetPassword;
    property Exclusive: Boolean read FExclusive write SetExclusive default false;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly default false;
    property TableCharSet: TCharSet read FCharSet write SetCharSet default csOEMCharSet;
    property AutoDateFields: Boolean read FAutoDate write FAutoDate default true;
    property AutoTimeFields: Boolean read FAutoTime write FAutoTime default true;
    property FieldDefs stored FieldDefsStored;
    property StoreDefs: Boolean read FStoreDefs write FStoreDefs default False;
    property IndexDefs: TIndexDefs read FIndexDefs write SetIndexDefs stored IndexDefsStored;
    property IndexFieldNames: string read GetIndexFieldNames write SetIndexFieldNames;
    property IndexName: string read GetIndexName write SetIndexName;
    property Active;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property BeforeRefresh;
    property AfterRefresh;
    property OnDeleteError;
    property OnEditError;
    property OnNewRecord;
    property OnPostError;
    property OnCalcFields;
    property AutoCalcFields;
    property Filtered;
    property OnFilterRecord;
//    property Filter;
//    property FilterOptions;
  end;

  TClarionMemoStream = class(TStream)
    FField: TBLOBField;
    FDataSet: TClarionTable;
    FMode: TBlobStreamMode;
    FPos: Longint;
    FRec: PDataHeader;
    function GetBlobSize: Longint;
  public
    constructor Create(Field: TBLOBField; Mode: TBlobStreamMode);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

resourcestring
  erCanWriteMemo = 'Can not write memo to deleted record';
  erNoTableName = 'No table name';
  erCompressedFile = 'The compressed files are not supported';
  erInvalidMemo = 'Invalid Memo file';
  erNoRecords = 'No records';
  erStrangeEncrypt = 'Strange encryptyon';
  erInvalidVersion = 'Invalid version';
  erNoOpenedRead = 'Not opened for read';
  erNoOpenedWrite = 'Not opened for write';
  erUnsuppType = 'Unsupported field type';
  erMemoExist = 'Memo already exist';
  erNoExlusive = 'Non exclusive access';
  erInvalidPassword = 'Invalid password';
  erNoIndex = 'No active index';
  erInvalidKey = 'Missing or invalid key file';

IMPLEMENTATION

function PatchFieldName(const S: String): String;
var p: integer;
begin
  p:=pos(':', S);
  if p>0 then Result := Trim(Copy(S, p+1, Length(S)))
   else Result := Trim(S);
end;

procedure DecodeBuffer(var Buf; BufSize, Id: Word);
var i: integer;
begin
  for i:= 0 to BufSize div 2 - 1 do
   TWordArray(Buf)[i] := TWordArray(Buf)[i] XOR Id
end;

constructor TClarionTable.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FIndexDefs:=TIndexDefs.Create(Self);
  FOpenMode:=fmOpenReadWrite;
  FShareMode:=fmShareDenyNone;
  FCharSet:=csOEMCharSet;
  FAutoDate:=true;
  FAutoTime:=true
end;

destructor TClarionTable.Destroy;
begin
  inherited;
  FIndexDefs.Free
end;

function KeyName(i: integer): string;
const Delta=ord('A')-ord('0')-10;
var j: integer;
begin
  Result:='.K00';
  Inc(i);
  j:=i mod 36; i:=i div 36;
  if j<=9 then inc(Result[4], j) else inc(Result[4], j + delta);
  if i<=9 then inc(Result[3], i) else inc(Result[3], i + delta)
end;

procedure TClarionTable.InternalOpen;
var i: integer;
    Mode: Word;
    s: String;
begin
  InternalOpenFile;
  FieldDefs.Updated:=False;
  FieldDefs.Update;
  IndexDefs.Updated:=False;
  IndexDefs.Update;
  BookmarkSize:=SizeOf(LongInt);
  if DefaultFields then CreateFields;
  BindFields(True);
  FBookmarkOfs:=FHeader.RecLen + CalcFieldsSize;
  FCurRec:=0;
  for i:=0 to High(FCDSKeys) do
   begin
     s:=ChangeFileExt(FTableName, KeyName(i));
     if FDatabaseName<>'' then s:=IncludeTrailingPathDelimiter(FDatabaseName)+s;
     Mode:=FShareMode;
     if ReadOnly then Mode:=Mode or fmOpenRead
      else Mode:=Mode or fmOpenReadWrite;
     FCDSKeys[i].Key:=TClarionKey.Create(s, Mode);
   end;
  SelectIndex;
  if FActiveKey<>nil then FActiveKey.Key.First;
  GetMem(FKeyBuffer, FHeader.RecLen);
  GetMem(FSaveKeyBuffer, FHeader.RecLen);
  ClearRecord(FKeyBuffer)
end;

procedure TClarionTable.InternalClose;
begin
  BindFields(False);
  if DefaultFields then DestroyFields;
  FActiveKey:=nil;
  InternalCloseFile;
  if FSaveKeyBuffer<>nil then FreeMem(FSaveKeyBuffer);
  if FKeyBuffer<>nil then FreeMem(FKeyBuffer)
end;

procedure TClarionTable.InternalOpenFile;
 var i, j: Integer;
     FN: String;
 begin
   FReadOnly:=FOpenMode=fmOpenRead;
   if FTableName = '' then DatabaseError(erNoTableName, Self);
   FN:=FTableName;
   if ExtractFileExt(FN)='' then FN:=ChangeFileExt(FN,'.DAT');
   if FDatabaseName<>'' then FN:=IncludeTrailingPathDelimiter(FDatabaseName)+FN;
   try
     FFile:=TFileStream.Create(FN, FShareMode or FOpenMode)
   except
    on EFOpenError do
     if FReadOnly then raise else
       begin
         FFile:=TFileStream.Create(FN, FShareMode or fmOpenRead);
         FReadOnly:=true
       end
   end;
   InternalReadHeader;
   FFileUpdated:=False;

   if (FHeader.SFAtr AND SIGN_COMPRESSED) <> 0 then
     DatabaseError(erCompressedFile);

   if (FHeader.SFAtr AND SIGN_MEMO) <> 0 then
    begin
     if FReadOnly then i:=fmOpenRead else i:=fmOpenReadWrite;
     FMemo:=TFileStream.Create(ChangeFileExt(FTableName,'.MEM'),FShareMode or i);
     FMemo.ReadBuffer(FMemoHeader, SizeOf(FMemoHeader));
     if FMemoHeader.MemoSIG <> MEMO_21_SIG then DatabaseError(erInvalidMemo)
    end;

   SetLength(FCDSFields, FHeader.NumFlds);
   for i:=0 to FHeader.NumFlds-1 do
    begin
      FFile.ReadBuffer(FCDSFields[i], sizeof(TFieldRecord));
      if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
        DecodeBuffer(FCDSFields[i], sizeof(TFieldRecord), FId);
    end;

   // Keys
   SetLength(FCDSKeys, FHeader.NumKeys);
   for i:=0 to FHeader.NumKeys-1 do with FCDSKeys[i] do
    begin
      FFile.ReadBuffer(FKeyRecord, sizeof(TKeyRecord));
      if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
       DecodeBuffer(FKeyRecord, sizeof(TKeyRecord), FId);

      Initialize(FKeyItems);
      SetLength(FKeyItems, FKeyRecord.NumComps);
      for j:=0 to FKeyRecord.NumComps-1 do
       begin
         FFile.ReadBuffer(FKeyItems[j], SizeOf(TKeyItem));
         if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
           DecodeBuffer(FKeyItems[j], SizeOf(TKeyItem), FId)
       end
    end;

   // Pictures
   SetLength(FCDSPictures, FHeader.NumPics);
   if FHeader.NumPics > 0 then
    for i:=0 to FHeader.NumPics-1 do with FCDSPictures[i] do
    begin
      FFile.ReadBuffer(PicLen, 2);
      Initialize(PicStr);
      SetLength(PicStr, PicLen);
      FFile.ReadBuffer(PicStr[1], PicLen);
      if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
        DecodeBuffer(PicStr[1], PicLen, FId)
    end;

   // Arrays
   SetLength(FCDSArrays, FHeader.NumArrs);
   for i:=0 to FHeader.NumArrs-1 do with FCDSArrays[i] do
    begin
      FFile.ReadBuffer(FArrayRecord, sizeof(TArrayRecord));
{      if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
        DecodeBuffer(FArrayRecord, sizeof(TArrayRecord), FId);}
      Initialize(FArrayItems);
      SetLength(FArrayItems, FArrayRecord.TotDim);
      for j:=0 to FArrayRecord.TotDim-1 do
        begin
          FFile.ReadBuffer(FArrayItems[j], sizeof(TArrayItem));
{          if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
            DecodeBuffer(FArrayItems[j], sizeof(TArrayItem), FId);}
        end;
    end
 end;

procedure TClarionTable.InternalCloseFile;
 var i: Integer;
 begin
   FReadOnly:= FOpenMode=fmOpenRead;
   FreeAndNil(FMemo);
   FreeAndNil(FFile);
   FCDSFields:=nil;

   for i:=0 to High(FCDSPictures) do Finalize(FCDSPictures[i].PicStr);
   FCDSPictures:=nil;

   for i:=0 to High(FCDSArrays) do FCDSArrays[i].FArrayItems:=nil;
   FCDSArrays:=nil;

   for i:=0 to High(FCDSKeys) do with FCDSKeys[i] do
    begin
      Key.Free;
      FKeyItems:=nil
    end;
   FCDSKeys:=nil
 end;

procedure TClarionTable.ReadRecord(RecNum:longint; var Buffer; Len: Word);
begin
  FFile.Seek(FHeader.Offset + FHeader.RecLen * (RecNum-1), soFromBeginning);
  if Len=0 then len:=FHeader.RecLen;
  FFile.ReadBuffer(Buffer, Len)
end;

procedure TClarionTable.WriteRecord(RecNum: longint; const Buffer; Len: Word);
begin
  FFile.Seek(FHeader.Offset + FHeader.RecLen * (RecNum-1), soFromBeginning);
  if Len=0 then len:=FHeader.RecLen;
  FFile.WriteBuffer(Buffer, Len)
end;

function TClarionTable.Translate(Src, Dest: PChar; ToOem: Boolean): Integer;
begin
  if Src = nil then begin Result:=0; exit end;
  Result := StrLen(Src);
  if FCharSet=csOEMCharSet then
   if ToOem then CharToOemBuff(Src, Dest, Result+1)
   else OemToCharBuff(Src, Dest, Result+1) else
    if (Src <> Dest) then Move(Src^, Dest^, Result+1);
end;

function TClarionTable.IsCursorOpen: Boolean;
 begin
   Result:=Assigned(FFile) AND (TFileStream(FFile).Handle > 0)
 end;

function Contain(const s, c: string): boolean;
 begin
   Contain:= pos(c, Uppercase(S))<>0
 end;

procedure TClarionTable.InternalInitFieldDefs;
var i: integer;
    fl: PFieldRecord;
    fOpened: boolean;
begin
  fOpened:=IsCursorOpen;
  if not fOpened then InternalOpenFile;
  Fielddefs.Clear;
  for i := 0 to FHeader.NumFlds - 1 do with FieldDefs.AddFieldDef do
   begin
     fl:=@FCDSFields[i];
     Name:=PatchFieldName(fl^.FldName);
     FieldNo:=i+1;
     Precision:=0;
     Size := 0;
     case fl^.FldType of
       FLD_BYTE, FLD_SHORT: DataType:= ftSmallInt;
       FLD_LONG:
        begin
          DataType:=ftInteger;

          if AutoDateFields and(DataType = ftInteger) then
            if Contain(Name, 'DAT') then DataType:= ftDate;
          if AutoTimeFields and(DataType = ftInteger) then
            if Contain(Name, 'TIM') then DataType:= ftTime;
        end;
       FLD_REAL:
         begin
           DataType:= ftFloat;
           Precision:= fl^.DecDec;
         end;
       FLD_DECIMAL:
         begin
           DataType:= ftBCD;
           Precision:= fl^.DecSig + fl^.DecDec;
           Size:= fl^.DecDec;
         end;
       FLD_STRING,
       FLD_PICTURE:
         begin
           DataType:= ftString;
           Size:= fl^.Length;
         end;
       FLD_GROUP: DataType := ftUnknown;
     end; // case
     Required:= False;
   end; // for-with

  if FHeader.SFAtr AND SIGN_MEMO <> 0 then with FieldDefs.AddFieldDef do
   begin
     DataType := ftMemo;
     Name := Trim(FHeader.MemName);
     Size := FHeader.MemoLen
   end;

  if not fOpened then InternalCloseFile
end;

function TClarionTable.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;

 function ResyncKey: Boolean;
  begin
    Result:=GetRecord(Buffer, gmCurrent, false) = grOk;
    if Result then
     Result:=FActiveKey.Key.FindRec(FCurRec, PChar(CreateKey(Buffer)))
  end;

 var FirstRec, LastRec: LongInt;
     Encripted: Boolean;
 begin
   if not Exclusive then InternalReadHeader;
   with FHeader do if FShowDeleted then
    begin
      LastRec:=NumRecs; FirstRec:=1
    end else
    begin
      LastRec:=LogEOF; FirstRec:=LogBOF
    end;
   Result:=grOK;
   repeat
     case GetMode of
      gmNext:
        if FActiveKey<>nil then with FActiveKey.Key do
         begin
           if FCurRec<FirstRec then FCurRec:=First else
           if FCurRec>=LastRec then Result:=grEOF else
            begin
              if CurRec<>FCurRec then if not ResyncKey then Result:=grError;
              FCurRec:=Next;
              if FCurRec = 0 then Result:=grEOF
            end;
         end else
        if FCurRec >= LastRec then Result:=grEOF else
         if FCurRec<FirstRec then FCurRec:=FirstRec else Inc(FCurRec);

      gmPrior:
        if FActiveKey<>nil then with FActiveKey.Key do
         begin
           if FCurRec>LastRec then FCurRec:=Last else
           if FCurRec<=FirstRec then Result:=grBOF else
            begin
              if CurRec<>FCurRec then if not ResyncKey then Result:=grError;
              FCurRec:=Prev;
              if FCurRec = 0 then Result:=grBOF
            end;
         end else
        if FCurRec <= FirstRec then Result:=grBOF else
         if FCurRec > LastRec then FCurRec:=LastRec else Dec(FCurRec)

      else if (FCurRec < FirstRec)or(FCurRec > LastRec) then Result:=grError;
     end;

     Encripted:= FHeader.SFAtr AND SIGN_ENCRYPTED <> 0;
     case Result of
      grError: if DoCheck then DatabaseError(erNoRecords, Self) else exit;
      grOK:
       begin
         ReadRecord(FCurRec, Buffer^);
         if not ShowDeleted then
          if (PDataHeader(Buffer)^.RHd and DATA_DEL)<>0 then
           begin
             if GetMode=gmCurrent then GetMode:=gmNext;
             Continue
           end;
         if FFilterOn then
          begin
            if Encripted then
             begin
               DecodeBuffer(Buffer[SizeOf(TDataHeader)],
                FHeader.RecLen-SizeOf(TDataHeader), FId);
               Encripted:=False
             end;
            if not FilterRecord(Buffer) then
              begin
                if GetMode=gmCurrent then GetMode:=gmNext;
                Continue
              end;
          end;
         break
       end
      else exit
     end //case
   until false;

   if Encripted then
    DecodeBuffer(Buffer[SizeOf(TDataHeader)],
     FHeader.RecLen-SizeOf(TDataHeader), FId);
   GetCalcFields(Buffer);
   with PRecInfo(Buffer+FBookmarkOfs)^ do
    begin
      BookmarkFlag:= bfCurrent;
      Bookmark:= FCurRec;
      Initialize(MemoBuf);
      OldMemoPtr:= PDataHeader(Buffer)^.RPtr
    end
 end;

procedure TClarionTable.InternalReadHeader;
 procedure Crack;
 begin
   if not IsTableHaveMemo then
    begin
      FId := Swap(FHeader.MemoLen);
      if CheckHeader then Exit;
      FId := Swap(PWordArray(@FHeader.MemName)[5] XOR $2020);
      if CheckHeader then Exit;
    end;
   FId := Swap(FHeader.Offset SHR 16);
   if CheckHeader then Exit;
   FId := Swap(PWordArray(@FHeader.RecName)[5] XOR $2020);
   if CheckHeader then Exit;
   FId := Swap(PWordArray(@FHeader.MemName)[5] XOR $2020);
   if CheckHeader then Exit;
   Close;
   DatabaseError(erStrangeEncrypt)
 end;

var i: Integer;
begin
  FFile.Seek(0, soFromBeginning);
  FFile.ReadBuffer(FHeader, SizeOf(THeader));
  if FHeader.FileSIG <> VERSION_21_SIG then DataBaseError(erInvalidVersion);

  if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
   if not CheckHeader then
    if Password = ' ' then Crack else DataBaseError(erInvalidPassword);

  FFileUpdated:=(FLastDate<>FHeader.ChgDate)or(FLastTime<>FHeader.ChgTime);
  FLastDate:=FHeader.ChgDate; FLastTime:=FHeader.ChgTime;

  if FFileUpdated then
   for i:=0 to High(FCDSKeys) do FCDSKeys[i].Key.UpdateHeader
end;

procedure TClarionTable.InternalWriteHeader;
begin
  FHeader.ChgDate:= Round(Date)+DELTA_DAYS;
  FHeader.ChgTime:= Round(Time*360000*24)+1;
  if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
   begin
     FHeader.CheckSum:= 0;
     FHeader.CheckSum:= CalcCheckSum;
     DecodeHeader
   end;
  FFile.Seek(0, soFromBeginning);
  FFile.WriteBuffer(FHeader, SizeOf(FHeader));
  if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then DecodeHeader;
  FLastDate:=FHeader.ChgDate; FLastTime:=FHeader.ChgTime;
  FFileUpdated:=false
end;

procedure TClarionTable.ClearRecord(Buffer: PChar);
var i: integer;
begin
  FillChar(Buffer^, FHeader.RecLen, 0);
  for i:=0 to High(FCDSFields) do with FCDSFields[i] do
   case FldType of
    FLD_STRING,
    FLD_PICTURE: FillChar(Buffer[FOffset + Sizeof(TDataHeader)], Length, $20)
   end
end;

procedure TClarionTable.InternalInitRecord(Buffer: PChar);
begin
  ClearRecord(Buffer);
  FillChar(Buffer[FBookmarkOfs], SizeOf(TRecInfo), 0)
end;

procedure TClarionTable.ClearCalcFields(Buffer: PChar);
begin
  FillChar(Buffer[FHeader.RecLen], CalcFieldsSize, 0)
end;

function TClarionTable.GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean;
var
  SrcBuf: PChar;
  PCurField: PFieldRecord;
  Field: TField;
  L: integer;
begin
  if (State = dsBrowse) and IsEmpty then begin Result:=False; Exit end;
  Field:=FieldByNumber(FieldNo);
  if Field.IsBlob then with PDataHeader(CurrentBuffer)^ do
   begin
     Result:= (RPtr<>0)and(RHd and DATA_DEL =0);
     exit
   end;
  if FieldNo<=0 then begin Result:=False; exit end;

  PCurField:= @FCDSFields[FieldNo-1];
  SrcBuf:= CurrentBuffer + PCurField^.FOffset + Sizeof(TDataHeader);
  Result:= True;

  if Buffer = nil then
   begin
     if (Field.DataType=ftDate)and(PLongInt(SrcBuf)^ =0) then Result:=false;
     exit
   end;

  case PCurField^.FldType of
    FLD_BYTE, FLD_SHORT, FLD_REAL:
      Move(SrcBuf^, Buffer^, PCurField^.Length);
    FLD_LONG:
      case FieldByNumber(FieldNo).DataType of
       ftDate:
         if PLongInt(SrcBuf)^ <> 0
          then PLongInt(Buffer)^:= PLongInt(SrcBuf)^ + DELTA_DAYS2
          else begin PLongInt(Buffer)^:=0; Result:=false end;
       ftTime:
         if PLongInt(SrcBuf)^ <> 0
          then PLongInt(Buffer)^:= (PLongInt(SrcBuf)^ -1)* 10
          else begin PLongInt(Buffer)^:=0; Result:=false end;
       else Move(SrcBuf^, Buffer^, PCurField^.Length)
      end;
    FLD_DECIMAL:
      with PBCD(Buffer)^ do
       begin
         Precision := PCurField^.DecSig +PCurField^.DecDec + 1;
         SignSpecialPlaces:= PCurField^.DecDec;
         if PByte(SrcBuf)^ > $0F then SignSpecialPlaces:= SignSpecialPlaces OR $80;
         Move(SrcBuf^, Fraction, PCurField^.Length);
         if Byte(Fraction[0]) > 0 then
           Byte(Fraction[0]):= Byte(Fraction[0]) AND $0F
       end;
    FLD_STRING, FLD_PICTURE:
      begin
        FillChar(Buffer^, Field.DataSize, 0);
        Move(SrcBuf^, Buffer^, PCurField^.Length);
        L:= PCurField^.Length-1;
        while (L>=0) and (PChar(Buffer)[L]<=' ') do dec(L); // trim string field
        PChar(Buffer)[L+1]:=#0;
      end;
    FLD_GROUP:
      begin
        Result:= false
      end
  end // case
end;

function TClarionTable.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var RecBuf: PChar;
begin
  if Field.FieldNo > 0 then
   begin
    Result := GetFieldData(Field.FieldNo, Buffer);
    exit
   end;

  Result := True;
  if State in [dsBrowse, dsEdit, dsInsert, dsCalcFields, dsBlockRead] then
   begin
     if State=dsBlockRead then RecBuf:=TempBuffer else RecBuf:=CurrentBuffer;
     Inc(RecBuf, FHeader.RecLen + Field.Offset);
     Result := Boolean(RecBuf[0]);
     if Result and (Buffer <> nil) then Move(RecBuf[1], Buffer^, Field.DataSize)
   end
end;

procedure TClarionTable.SetFieldData(Field: TField; Buffer: Pointer);
var DestBuf : PChar;
    Len, LenS: integer;
    FType: Byte;
    C: Char;
begin
  with Field do
  begin
    if IsBlob then exit;
    if not (State in dsWriteModes) then DatabaseError(SNotEditing, Self);
    if (State = dsSetKey) and ((FieldNo < 0) or not IsIndexField)
      then DatabaseErrorFmt(SNotIndexField, [DisplayName]);
    DestBuf:=CurrentBuffer;
    if FieldNo > 0 then
    begin
      if State = dsCalcFields then DatabaseError(SNotEditing, Self);
      if ReadOnly and not (State in [dsSetKey, dsFilter]) then
        DatabaseErrorFmt(SFieldReadOnly, [DisplayName]);
      Validate(Buffer);
      if FieldKind <> fkInternalCalc then
       begin
         with FCDSFields[FieldNo-1] do
          begin
            Len:=Length;
            FType:=FldType;
            Inc(DestBuf, FOffset + Sizeof(TDataHeader))
          end;
         if Buffer=nil then
          begin
            if FType=FLD_STRING then C:=#32 else C:=#0;
            FillChar(DestBuf^, Len, C)
          end else
          case FType of
           FLD_BYTE, FLD_SHORT, FLD_REAL:
               Move(Buffer^, DestBuf^, Len);
           FLD_LONG:
             case Field.DataType of
              ftDate:
                begin
                  PLongInt(DestBuf)^:= PLongInt(Buffer)^;
                  if PLongInt(DestBuf)^<>0
                    then dec(PLongInt(DestBuf)^, DELTA_DAYS2)
                end;
              ftTime:
                if PLongInt(Buffer)^ = 0 then PLongInt(DestBuf)^:=0
                 else PLongInt(DestBuf)^:= PLongInt(Buffer)^ div 10 +1;
              else Move(Buffer^, DestBuf^, Len)
             end;
           FLD_DECIMAL:
             with PBCD(Buffer)^ do
             begin
               Move(Fraction[(Precision div 2) - Len], DestBuf^, Len);
               if Boolean(SignSpecialPlaces AND $80) then
                 Byte(DestBuf^) := Byte(DestBuf^) OR $F0;
             end;
           FLD_STRING, FLD_PICTURE:
             begin
               LenS:=StrLen(Buffer);
               if LenS>Len then LenS:=Len;
               Move(Buffer^, DestBuf^, LenS);
               FillChar(DestBuf[Lens], Len-Lens, $20)
             end
          end // case
       end
    end else {fkCalculated, fkLookup}
    begin
      Inc(DestBuf, FHeader.RecLen+Offset);
      Boolean(DestBuf[0]) := Buffer<>nil;
      if Boolean(DestBuf[0]) then Move(Buffer^, DestBuf[1], DataSize)
    end;
  end; // with
  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
    DataEvent(deFieldChange, Longint(Field))
end;

procedure TClarionTable.InternalHandleException;
begin
  ApplicationHandleException(Self)
end;

function TClarionTable.GetCanModify: Boolean;
 begin
   Result:={(FHeader.SFAtr AND SIGN_READONLY = 0) and} not FReadOnly
 end;

function TClarionTable.GetRecordSize: Word;
 begin
   Result:=FHeader.RecLen
 end;

function TClarionTable.CurrentBuffer: PChar;
begin
  case State of
   dsCalcFields: Result:= CalcBuffer;
   dsFilter: Result:= FFilterBuffer;
   dsSetKey: Result:= FKeyBuffer;
   else Result:= ActiveBuffer
  end;
end;

function TClarionTable.AllocRecordBuffer: PChar;
 begin
   GetMem(Result, FBookmarkOfs + SizeOf(TRecInfo))
 end;

procedure TClarionTable.FreeRecordBuffer(var Buffer: PChar);
 begin
   FreeMem(Buffer{, FBookmarkOfs + SizeOf(TRecInfo)})
 end;

function TClarionTable.GetRecordCount: Longint;
begin
  Result := FHeader.NumRecs;
  if not FShowDeleted then dec(Result, FHeader.NumDels)
end;

function TClarionTable.GetRecNo: LongInt;
begin
  CheckActive;
  UpdateCursorPos;
  if (FCurRec = 0) and (RecordCount > 0) then Result:=1
   else Result := FCurRec
end;

procedure TClarionTable.SetRecNo(Value : Integer);
begin
  CheckActive;
  if (Value > 0) and (Value <= FHeader.NumRecs) then
   begin
     FCurRec := Value;
     Resync([])
   end
end;

procedure TClarionTable.InternalSetToRecord(Buffer: PChar);
begin
  FCurRec:=PRecInfo(Buffer + FBookmarkOfs)^.Bookmark;
end;

procedure TClarionTable.InternalFirst;
 begin
   FCurRec := 0
 end;

procedure TClarionTable.InternalLast;
 begin
   FCurRec := FHeader.NumRecs+1
 end;

procedure TClarionTable.InternalRefresh;
 begin
   if not FExclusive then InternalReadHeader
 end;

procedure TClarionTable.InternalPost;
begin
  inherited;
  RecordPostAppend(PDataHeader(ActiveBuffer), State = dsEdit)
end;

procedure TClarionTable.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  RecordPostAppend(PDataHeader(Buffer), False)
end;

procedure TClarionTable.RecordPostAppend(RecBuffer: PDataHeader; EditMode: boolean);

 procedure UpdateMemo;
  var NewChunk: Longint;
      FromDeleted: boolean;
      n, l, cn, ps: Longint;
      Chunk: TMemoData;
  const BlockSize=SizeOf(Chunk.Memo);

  function GetNewChunk: Longint;
   begin
     if (NewChunk=0)and not FromDeleted and not FExclusive then
      begin
        FMemo.Seek(2, soFromBeginning);
        FMemo.ReadBuffer(FMemoHeader.FirstDel, 4)
      end;
     if FMemoHeader.FirstDel<>0 then
      begin
        Result:=FMemoHeader.FirstDel;
        FromDeleted:=true
      end else
      begin
        if NewChunk<>0 then inc(NewChunk) else
          NewChunk:=(FMemo.Size+2*256-7)div 256;
        Result:=NewChunk
      end
   end;  //GetNewChunk

  begin  //UpdateMemo
    FromDeleted:=false;
    NewChunk:=0;
    ps:=0;
    n:=FHeader.MemoLen;
    with PRecInfo(PChar(RecBuffer)+FBookmarkOfs)^ do
     repeat
       if n=0 then
        begin
          DeleteMemo(RecBuffer^.RPtr);
          RecBuffer^.RPtr:=0;
          exit
        end;
       if MemoBuf[n-1]=' ' then dec(n) else break
     until false;
    if RecBuffer^.RPtr=0 then RecBuffer^.RPtr:=GetNewChunk;
    cn:=RecBuffer^.RPtr;
    repeat
      if NewChunk<>0 then Chunk.NextBlock:=0 else
       begin
         ReadMemoChunk(cn, Chunk, SizeOf(Chunk.NextBlock));
         if FromDeleted then FMemoHeader.FirstDel:=Chunk.NextBlock
       end;
      if n<BlockSize then l:=n else l:=BlockSize;
      Move(PRecInfo(PChar(RecBuffer)+FBookmarkOfs)^.MemoBuf[ps],Chunk.Memo,l);
      dec(n, l); inc(ps, l);
      if l<BlockSize then FillChar(Chunk.Memo[l], BlockSize-l, $20);
      if (n<>0)and(Chunk.NextBlock=0) then Chunk.NextBlock:=GetNewChunk;
      if (n=0)and(Chunk.NextBlock<>0) then
       begin
         if not FromDeleted then DeleteMemo(Chunk.NextBlock);
         Chunk.NextBlock:=0
       end;
      WriteMemoChunk(cn, Chunk);
      cn:=Chunk.NextBlock
    until n=0;
    if FromDeleted then
     begin
       FMemo.Seek(2,soFromBeginning);
       FMemo.WriteBuffer(FMemoHeader.FirstDel, 4)
     end
  end; //UpdateMemo

var RH: TDataHeader;
begin
  if not Exclusive then InternalReadHeader;
  if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
    DecodeBuffer(PChar(RecBuffer)[SizeOf(TDataHeader)], FHeader.RecLen - SizeOf(TDataHeader), FId);
  if EditMode then
   begin
     InternalSetToRecord(PChar(RecBuffer));
     RecBuffer^.RHd := RecBuffer^.RHd or DATA_REV;
   end else
   begin
     if (FHeader.NumDels = 0) or ((FHeader.SFAtr AND SIGN_RECLAIM) = 0) then
      begin
        Inc(FHeader.NumRecs);
        FCurRec:=FHeader.NumRecs
      end else
      begin
        FCurRec:=FHeader.FreeRec;
        ReadRecord(FCurRec, RH, SizeOf(RH));
        FHeader.FreeRec:=RH.RPtr;
        Dec(FHeader.NumDels)
      end;
     RecBuffer^.RHd:=DATA_NEW;
     if FHeader.LogBOF > FCurRec then FHeader.LogBOF:=FCurRec;
     if FHeader.LogEOF < FCurRec then FHeader.LogEOF:=FCurRec;
   end;

  if (FHeader.SFAtr AND SIGN_MEMO <> 0) and
     (RecBuffer^.RHd AND DATA_DEL = 0) then
    with PRecInfo(PChar(RecBuffer)+FBookmarkOfs)^ do
     begin
       if (RecBuffer^.RPtr=0) then DeleteMemo(OldMemoPtr) else
        begin
          RecBuffer^.RPtr:=OldMemoPtr;
          if Updated then UpdateMemo
        end;
       Updated:=false;
       MemoBuf:=nil
     end;

  WriteRecord(FCurRec, RecBuffer^);
  InternalWriteHeader
end;

procedure TClarionTable.InternalDelete;

 procedure FindLogicalBOF;
  var B: Byte;
      n: LongInt;
  begin
    for n:=FHeader.LogBOF+1 to FHeader.LogEOF do
     begin
       ReadRecord(n, B, SizeOf(B));
       if (B and DATA_DEL) = 0 then
        begin
          FHeader.LogBOF:=n;
          Exit
        end
     end;
    FHeader.LogEOF:=0;
    FHeader.LogBOF:=FHeader.NumRecs
  end;

 procedure FindLogicalEOF;
  var B: Byte;
      n: LongInt;
  begin
    for n:=FHeader.LogEOF-1 downto FHeader.LogBOF do
     begin
       ReadRecord(n, B, SizeOf(B));
       if (B and DATA_DEL) = 0 then
        begin
          FHeader.LogEOF:=n;
          Exit
        end
     end;
    FHeader.LogEOF:=0;
    FHeader.LogBOF:=FHeader.NumRecs
  end;

var RecBuffer: PDataHeader;
begin
  if not FExclusive then InternalReadHeader;
//  UpdateCursorPos;
  RecBuffer := PDataHeader(ActiveBuffer);
  if (RecBuffer^.RHd and DATA_DEL) <> 0 then exit;
  RecBuffer^.RHd:= RecBuffer^.RHd OR DATA_DEL;
  RecBuffer^.RPtr:= FHeader.FreeRec;
  FHeader.FreeRec:= FCurRec;
  Inc(FHeader.NumDels);
  WriteRecord(FCurRec, RecBuffer^, SizeOf(RecBuffer^));
  if FHeader.LogBOF = FCurRec then FindLogicalBOF else
   if FHeader.LogEOF = FCurRec then FindLogicalEOF;
  InternalWriteHeader;
  if (FHeader.SFAtr AND SIGN_MEMO) <> 0 then
    DeleteMemo(PRecInfo(PChar(RecBuffer)+FBookmarkOfs)^.OldMemoPtr);
end;

procedure TClarionTable.SetPassword(const APwd: String);
 Var i, k: Byte;
 begin
   CheckInactive;
   FPassword := APwd;
   k := Length(APwd);
   if k <= 0 then FId := 0 else
    begin
      k := k SHR 1;
      FId := $7F7F;
      for i:=1 to k do Inc(FId, PWord(@APwd[i])^)
    end
 end;

function TClarionTable.CalcCheckSum: Word;
var i: byte;
begin
  Result:=0;
  for i:=0 to SizeOf(THeader)-3 do Inc(Result, PByteArray(@FHeader)^[i]);
end;

procedure TClarionTable.DecodeHeader;
begin
  DecodeBuffer(FHeader.NumKeys, 80, FId)
end;

function TClarionTable.CheckHeader: Boolean;
var ChkSum : Word;
begin
  DecodeHeader;
  ChkSum:=FHeader.CheckSum;
  FHeader.CheckSum:= 0;
  Result:=True;
  if CalcCheckSum <> ChkSum then
   begin
    FHeader.CheckSum := ChkSum;
    DecodeHeader;
    Result:=False
   end
end;

function TClarionTable.IsRecordNew : Boolean;
begin
  Result := Byte(ActiveBuffer[0]) AND DATA_NEW <> 0
end;

function TClarionTable.IsRecordOld : Boolean;
begin
  Result := Byte(ActiveBuffer[0]) AND DATA_OLD <> 0
end;

function TClarionTable.IsRecordRevised : Boolean;
begin
  Result := Byte(ActiveBuffer[0]) AND DATA_REV <> 0
end;

function TClarionTable.IsRecordDeleted : Boolean;
begin
  Result := Byte(ActiveBuffer[0]) AND DATA_DEL <> 0
end;

function TClarionTable.IsRecordHeld : Boolean;
begin
  Result := Byte(ActiveBuffer[0]) AND DATA_HLD <> 0
end;

function TClarionTable.IsTableLocked : Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_LOCKED <> 0
end;

function TClarionTable.IsTableOwned : Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_OWNED <> 0
end;

function TClarionTable.IsTableEncrypted: Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_ENCRYPTED <> 0
end;

function TClarionTable.IsTableHaveMemo: Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_MEMO <> 0
end;

function TClarionTable.IsTableCompressed: Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_COMPRESSED <> 0
end;

function TClarionTable.IsTableReclaim: Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_RECLAIM <> 0
end;

function TClarionTable.IsTableReadonly: Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_READONLY <> 0
end;

function TClarionTable.IsTableCreated: Boolean;
begin
  Result := FHeader.SFAtr AND SIGN_CREATED <> 0
end;

procedure TClarionTable.SetCharSet(CharSet: TCharSet);
 begin
   FCharSet:=CharSet
 end;

procedure TClarionTable.SetDatabaseName(const Value: string);
begin
  if csReading in ComponentState then FDatabaseName := Value else
   if FDatabaseName <> Value then
    begin
      CheckInactive;
      FDatabaseName := Value;
      DataEvent(dePropertyChange, 0)
    end;
end;

procedure TClarionTable.SetTableName(const Value: TFileName);
begin
  if csReading in ComponentState then FTableName := Value else
  if (FTableName <> Value) then
  begin
    CheckInactive;
    FTableName := Value;
    DataEvent(dePropertyChange, 0);
  end;
end;

procedure TClarionTable.SetReadOnly(Value: Boolean);
 begin
   CheckInactive;
   FReadOnly:=Value;
   if Value then FOpenMode:=fmOpenRead else FOpenMode:=fmOpenReadWrite
 end;

procedure TClarionTable.SetExclusive(Value : Boolean);
begin
  CheckInactive;
  FExclusive := Value;
  if Value then FShareMode := fmShareExclusive
   else FShareMode := fmShareDenyNone;
end;
{
procedure TClarionTable.InternalEdit;
begin
// Get record -> activebuffer
end; }

{procedure TClarionTable.InternalInsert;
begin
end;}

procedure TClarionTable.InternalCancel;
begin
  with PRecInfo(ActiveBuffer+FBookmarkOfs)^ do
   begin
     Updated:=false;
     MemoBuf:=nil;
     PDataHeader(ActiveBuffer)^.RPtr:=OldMemoPtr
   end
end;

function TClarionTable.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PRecInfo(Buffer + FBookmarkOfs)^.BookmarkFlag
end;

procedure TClarionTable.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PRecInfo(Buffer + FBookmarkOfs)^.BookmarkFlag := Value
end;

procedure TClarionTable.InternalGotoBookmark(Bookmark: Pointer);
begin
  FCurRec := PLongInt(Bookmark)^
end;

procedure TClarionTable.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PLongInt(Data)^ := PRecInfo(Buffer + FBookmarkOfs).Bookmark
end;

procedure TClarionTable.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PRecInfo(Buffer + FBookmarkOfs).Bookmark := PLongInt(Data)^
end;

function TClarionTable.CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer;
begin
  if Bookmark1=nil then
   if Bookmark2=nil then Result := 0 else Result:=-1
  else if Bookmark2=nil then Result := 1
   else if PLongInt(Bookmark1)^ > PLongInt(Bookmark2)^ then Result := 1
    else if PLongInt(Bookmark1)^ < PLongInt(Bookmark2)^ then Result := -1
     else Result:=0
end;

function TClarionTable.BookmarkValid(Bookmark: TBookmark): Boolean;
var b: Byte;
begin
  Result:= False;
  if Bookmark = nil then exit;
  if not FExclusive then InternalReadHeader;
  if PLongInt(Bookmark)^ < 1 then exit;
  if PLongInt(Bookmark)^ > FHeader.NumRecs then exit;
  if not FShowDeleted then
   begin
     if PLongInt(Bookmark)^ < FHeader.LogBOF then exit;
     if PLongInt(Bookmark)^ > FHeader.LogEOF then exit;
     ReadRecord(PLongInt(Bookmark)^, b, SizeOf(b));
     if (b and DATA_DEL) <> 0 then exit;
   end;
  Result:= True
end;

procedure TClarionTable.EmptyTable;
var e: Boolean;
begin
  CheckInactive;
  e:=Exclusive;
  Exclusive:=true;
  try
    InternalOpenFile;
    FHeader.NumRecs:=0;
    FHeader.NumDels:=0;
    FHeader.FreeRec:=0;
    FHeader.LogEOF:=0;
    FHeader.LogBOF:=0;
    InternalWriteHeader;
    FFile.Size:=FHeader.Offset;
    if FHeader.SFAtr AND SIGN_MEMO <> 0 then
     begin
       FMemoHeader.FirstDel:=0;
       FMemo.Seek(0, soFromBeginning);
       FMemo.WriteBuffer(FMemoHeader, SizeOf(FMemoHeader));
       FMemo.Size:=SizeOf(FMemoHeader)
     end
  finally
    InternalCloseFile;
    Exclusive:=e
  end
end;

function TClarionTable.IsSequenced: Boolean;
begin
  Result:= (FActiveKey = nil) and not Filtered
end;

(* TClarion memo implementation *)

function TClarionTable.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
  Result:= TClarionMemoStream.Create(Field as TBLOBField, Mode)
end;

procedure TClarionTable.ReadMemoChunk(RPtr: Integer; var Chunk; Size: Word);
const HeaderSize=4;
begin
  FMemo.Seek((RPtr - 1) SHL 8 + 6, soFromBeginning);
  FMemo.ReadBuffer(Chunk, Size);
  if (FHeader.SFAtr AND SIGN_ENCRYPTED) <> 0 then
   DecodeBuffer(PChar(@Chunk)[HeaderSize], Size-HeaderSize, FId)
end;

procedure TClarionTable.WriteMemoChunk(RPtr: Integer; var Chunk; Size: Word);
const HeaderSize=4;
begin
  if (FHeader.SFAtr AND SIGN_ENCRYPTED) <> 0 then
   DecodeBuffer(PChar(@Chunk)[HeaderSize], Size-HeaderSize, FId);
  FMemo.Seek((RPtr - 1) SHL 8 + 6, soFromBeginning);
  FMemo.WriteBuffer(Chunk, Size)
end;

procedure TClarionTable.DeleteMemo(Ptr: LongInt);
var P, T: LongInt;
begin
  if Ptr=0 then Exit;
  if not FExclusive then
   begin
     FMemo.Seek(2, soFromBeginning);
     FMemo.ReadBuffer(FMemoHeader.FirstDel, 4)
   end;
  P:=FMemoHeader.FirstDel;
  FMemoHeader.FirstDel:=Ptr;
  FMemo.Seek(2, soFromBeginning);
  FMemo.WriteBuffer(FMemoHeader.FirstDel, 4);
  if P=0 then exit;
  repeat
    ReadMemoChunk(Ptr, T, SizeOf(T));
    if (T=0)or(T=Ptr) then break;
    Ptr:=T
  until false;
  WriteMemoChunk(Ptr, P, SizeOf(P))
end;

procedure TClarionTable.PrepareMemoBuffer;
const BlockSize=252;
var i, n, t: Integer;
    P: Longint;
    M: TMemoData;
    RHdr: PDataHeader;
begin
  RHdr:=PDataHeader(ActiveBuffer);
  with PRecInfo(PChar(RHdr)+FBookmarkOfs)^ do
   begin
     if Length(MemoBuf)<>0 then exit;
     P:=RHdr^.RPtr;
     Updated:=false;
     n:=FHeader.MemoLen;
     SetLength(MemoBuf, n+1);
     i:=0;
     while (P<>0)and(n>0) do
      begin
        ReadMemoChunk(P, M);
        P:=M.NextBlock;
        if n<SizeOf(M.Memo) then t:=n else t:=SizeOf(M.Memo);
        Move(M.Memo, MemoBuf[i], t);
        inc(i, t); dec(n, t)
      end;
     if n>0 then FillChar(MemoBuf[i], n, $20)
   end
end;

(* TClarionMemoStream *)

constructor TClarionMemoStream.Create(Field: TBLOBField; Mode: TBlobStreamMode);
begin
  inherited Create;
  FMode:= Mode;
  FField:= Field;
  FDataSet:= Field.DataSet as TClarionTable;
  FRec:=PDataHeader(FDataSet.ActiveBuffer);
  if (FRec^.RHd and DATA_DEL <>0) and (Mode<>bmRead) then
   raise Exception.Create(erCanWriteMemo);

  if (Mode=bmWrite)and(FRec^.RPtr<>0) then
   with PRecInfo(PChar(FRec)+FDataSet.FBookmarkOfs)^ do
    begin
      Updated:=false;
      MemoBuf:=nil;
      FRec^.RPtr:=0;
      FDataSet.DataEvent(deFieldChange, Longint(FField))
    end;
  FPos:=0
end;

function TClarionMemoStream.Read(var Buffer; Count: Longint): Longint;
var t: array[0..1]of char;
begin
  if FMode=bmWrite then raise exception.create(erNoOpenedRead);
  Result:=GetBlobSize;
  if FPos>=Result then begin Result:=0; exit end;
  Dec(Result, FPos);
  if Result>Count then Result:=Count;
  Count:=Result;
  FDataset.PrepareMemoBuffer;
  Move(PRecInfo(PChar(FRec)+FDataSet.FBookmarkOfs)^.MemoBuf[FPos], Buffer, Count);
  Inc(FPos, Count);
  if not FField.Transliterate then Exit;
  dec(Count);
  t[0]:=PChar(@Buffer)[Count]; t[1]:=#0;
  PChar(@Buffer)[Count]:=#0;
  FDataSet.Translate(PChar(@Buffer), PChar(@Buffer), false);
  FDataSet.Translate(t, t, false);
  PChar(@Buffer)[Count]:=t[0]
end;

function TClarionMemoStream.Write(const Buffer; Count: Longint): Longint;
var t: char;
begin
  if FMode=bmRead then raise exception.create(erNoOpenedWrite);
  if (FPos+Count) > FDataSet.FHeader.MemoLen then
    Count:= FDataSet.FHeader.MemoLen - FPos;
  if Count<=0 then begin Result:=0; Exit end;
  Result:=Count;
  FDataset.PrepareMemoBuffer;
  FRec^.RPtr:=1; // Any non zero value for indication of presence memo
  with PRecInfo(PChar(FRec)+FDataSet.FBookmarkOfs)^ do
   begin
     Updated:=true;
     Move(Buffer, MemoBuf[FPos], Count);
     if FField.Transliterate then
      begin
        t:=MemoBuf[FPos+Count]; MemoBuf[FPos+Count]:=#0;
        FDataSet.Translate(PChar(@MemoBuf[FPos]), PChar(@MemoBuf[FPos]), True);
        MemoBuf[FPos+Count]:=t
      end;
     Inc(FPos, Count)
   end;
  FDataSet.DataEvent(deFieldChange, Longint(FField))
end;

function TClarionMemoStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
   soFromCurrent: Inc(FPos, Offset);
   soFromEnd: FPos:=GetBlobSize + Offset;
   else FPos:=Offset
  end;
  Result:=FPos
end;

function TClarionMemoStream.GetBlobSize: Longint;
begin
  if FRec^.RPtr=0 then Result:=0 else Result:= FDataSet.FHeader.MemoLen
end;

{ Design support}

procedure TClarionTable.DefChanged(Sender: TObject);
begin
  StoreDefs:=True
end;

function TClarionTable.FieldDefsStored: Boolean;
begin
  Result:=StoreDefs and (FieldDefs.Count > 0)
end;

function TClarionTable.IndexDefsStored: Boolean;
begin
  Result:=StoreDefs and (IndexDefs.Count > 0)
end;

procedure TClarionTable.SetIndexDefs(Value: TIndexDefs);
begin
  IndexDefs.Assign(Value)
end;

{--------------------------------------------------------------}

procedure StrToCharArray(const S: String; pC: array of char);
var L: integer;
begin
  L:=Length(S);
  if L>SizeOf(pC) then L:=SizeOf(pC);
  Move(S[1], pC, L);
  if L<SizeOf(pC) then FillChar(pC[L], SizeOf(pC)-L, ' ')
end;

procedure TClarionTable.FDT2FC(FD: TFieldDef; var FR: TFieldRecord);
begin
  StrToCharArray(FHeader.FilPrefx+':'+Uppercase(FD.Name), FR.FldName);
  with FR do
    begin
      FOffset:=FHeader.RecLen;
      case FD.DataType of
       ftString:
         begin
           fldtype:= FLD_STRING;
           Length:= FD.size
         end;
       ftInteger, ftDate, ftTime:
         begin
           fldtype:= FLD_LONG;
           Length:= 4
         end;
       ftSmallint:
         begin
           fldtype:= FLD_SHORT;
           Length:= 2
         end;
       ftBCD:
         begin
           fldtype:= FLD_DECIMAL;
           if FD.Precision<=0 then Length:=6 else Length:=FD.precision div 2+1;
           if Length>8 then Length:=8;
           decdec:= FD.size;
           if {(decdec<0) or} (decdec>8) then decdec:=8;
           decsig:= Length*2-1-decdec; //FD.precision;
         end;
       ftFloat, ftCurrency:
         begin
           fldtype := FLD_DECIMAL;
           Length  := 8;
           decdec  := 4;
           decsig  := 11; // Length*2-1-decdec;
         end;
       ftMemo, ftBlob:
         begin
           if FHeader.SFAtr AND SIGN_MEMO <> 0 then DatabaseError(erMemoExist);
           FHeader.SFAtr:=FHeader.SFAtr OR SIGN_MEMO;
           FHeader.MemoLen:=FD.size;
           StrToCharArray(Uppercase(FD.Name), FHeader.MemName)
         end
       else DatabaseError(erUnsuppType+FD.ClassName);
     end {case};
     Inc(FHeader.RecLen, Length)
   end
end;

procedure TClarionTable.CreateTable;
var Nam: Char3;
begin
  StrToCharArray(UpperCase(ExtractFileName(FTableName))+'000', Nam);
  CreateTable(Nam)
end;

procedure TClarionTable.CreateTable(const Prefix: Char3);
var i: integer;
begin
   CheckInActive;
   if not FExclusive then raise Exception.Create(erNoExlusive);
   try
      FCDSFields:=nil;
      FCDSKeys:=nil;
      FCDSPictures:=nil;
      FCDSArrays:=nil;
      FillChar(FHeader, SizeOf(FHeader), 0);
      with FHeader do
       begin
         filesig:=VERSION_21_SIG;
         NumFlds:=Fielddefs.Count;
         Offset:=SizeOf(FHeader)+NumFlds*SizeOf(TFieldRecord);
         RecName:='RECORD      ';
         FillChar(MemName, SizeOf(MemName), ' ');
         FilPrefx:=Prefix;
         RecPrefx:=Prefix
       end;
      SetLength(FCDSFields, FHeader.NumFlds);
      FillChar(FCDSFields[0], FHeader.NumFlds*SizeOf(FCDSFields[0]), 0);
      for i:=0 to FHeader.NumFlds-1 do FDT2FC(FieldDefs[i], FCDSFields[i]);
      Inc(FHeader.RecLen, SizeOf(TDataHeader));
      FFile:=TFileStream.Create(FTableName, fmCreate);
      InternalWriteHeader;
      for i:=0 to FHeader.NumFlds-1 do
        FFile.WriteBuffer(FCDSFields[i], SizeOf(TFieldRecord));
      if (FHeader.SFAtr AND SIGN_MEMO) <> 0 then
       begin
         FMemoHeader.MemoSIG:=MEMO_21_SIG;
         FMemo:=TFileStream.Create(ChangeFileExt(FTableName,'.MEM'), fmCreate);
         FMemo.WriteBuffer(FMemoHeader, 6)
       end
   finally
     InternalClose
   end
end;

{ Filtering }

procedure TClarionTable.SetFilterOptions(Value: TFilterOptions);
begin
  SetFilterData(Filter, Value)
end;

procedure TClarionTable.SetFilterText(const Value: string);
begin
  SetFilterData(Value, FilterOptions)
end;

procedure TClarionTable.SetFilterData(const Text: string; Options: TFilterOptions);
begin
  if Active then CheckBrowseMode;
  inherited SetFilterText(Text);
  inherited SetFilterOptions(Options);
  if Active and Filtered then First;
end;

procedure TClarionTable.SetFiltered(Value: Boolean);
begin
  inherited;
  FFilterOn:=Value;
  if Active then
  begin
    CheckBrowseMode;
    First
  end
end;

function TClarionTable.FilterRecord(Buf: PChar): boolean;
var
  SaveState: TDataSetState;
begin
  Result := True;
  SaveState := SetTempState(dsFilter);
  FFilterBuffer := Buf;
  try
    Result:=EvaluateFilter
  except
    ApplicationHandleException(Self)
  end;
  if Result and Assigned(OnFilterRecord) then
   try
     OnFilterRecord(Self, Result)
   except
     ApplicationHandleException(Self)
   end;
  RestoreState(SaveState)
end;

function TClarionTable.FindRecord(Restart, GoForward: Boolean): Boolean;
var Mode: TGetMode;
begin
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;
  FFilterOn:=True;
  try
    if GoForward then
    begin
      if Restart then InternalFirst;
      Mode:=gmNext
    end else
    begin
      if Restart then InternalLast;
      Mode:=gmPrior
    end;
    Result := GetRecord(TempBuffer, Mode, false) = grOk
  finally
    if not Filtered then FFilterOn:=False
  end;
  if Result then
   begin
     Resync([rmExact, rmCenter]);
     SetFound(True);
     DoAfterScroll
   end
end;

function TClarionTable.EvaluateFilter: boolean;
 begin
   Result:=True
 end;

{ Index Related }

procedure TClarionTable.Post;
begin
  if State = dsSetKey then PostKeyBuffer(True) else inherited
end;

procedure TClarionTable.Cancel;
begin
  if State = dsSetKey then PostKeyBuffer(False) else inherited
end;

procedure TClarionTable.PostKeyBuffer(Commit: Boolean);
begin
  DataEvent(deCheckBrowseMode, 0);
  if not Commit then Move(FSaveKeyBuffer^, FKeyBuffer^, FHeader.RecLen);
  SetState(dsBrowse);
  DataEvent(deDataSetChange, 0);
end;

procedure TClarionTable.UpdateIndexDefs;
var fOpened: Boolean;
    i, j: Integer;
begin
  fOpened:=IsCursorOpen;
  if not fOpened then InternalOpenFile;

  IndexDefs.Clear;
  for i:=0 to High(FCDSKeys) do with FCDSKeys[i], IndexDefs.AddIndexDef do
   begin
     Name:=PatchFieldName(FKeyRecord.KeyNams);
     Fields := '';
     for j:=0 to High(FKeyItems) do
      begin
        if j<>0 then Fields:=Fields+';';
        Fields:=Fields+PatchFieldName(FCDSFields[FKeyItems[j].FldNum-1].FldName)
      end;
     if FKeyRecord.CompType and KEY_DUPSW<>0 then Options:=Options+[ixUnique];
     if FKeyRecord.CompType and KEY_UPRSW<>0 then Options:=Options+[ixCaseInsensitive];
     if FKeyRecord.CompType and KEY_TYPMASK<>0 then Options:=Options+[ixNonMaintained];
   end;

  if not fOpened then InternalCloseFile
end;

function TClarionTable.GetIndexFieldNames: string;
begin
  if FFieldsIndex then Result := FIndexName else Result := ''
end;

function TClarionTable.GetIndexName: string;
begin
  if FFieldsIndex then Result := '' else Result := FIndexName;
end;

procedure TClarionTable.SetIndexFieldNames(const Value: string);
begin
  SetIndex(Value, Value <> '')
end;

procedure TClarionTable.SetIndexName(const Value: string);
begin
  SetIndex(Value, False)
end;

function TClarionTable.GetIsIndexField(Field: TField): Boolean;
var n, i: Integer;
begin
  Result:=False;
  if FActiveKey=nil then exit;
  n:=Field.FieldNo;
  with FActiveKey^ do
   for i:=0 to High(FKeyItems) do
    if FKeyItems[i].FldNum=n then
     begin
       Result:=True;
       Exit
     end
end;

procedure TClarionTable.SelectIndex;
var i: integer;
begin
  if FIndexName='' then begin FActiveKey:=nil; exit end;
  if not FFieldsIndex then i:=IndexDefs.IndexOf(FIndexName)else
  i:=IndexDefs.IndexOf(IndexDefs.FindIndexForFields(FIndexName).Name);
  if i<0 then DatabaseErrorFmt(SIndexDoesNotExist, [FIndexName], Self);
  FActiveKey:=@FCDSKeys[i]
end;

procedure TClarionTable.SetIndex(const Value: string; FieldsIndex: Boolean);
begin
  if Active then CheckBrowseMode;
  if (FIndexName <> Value) or (FFieldsIndex <> FieldsIndex) then
  begin
    FIndexName := Trim(Value);
    FFieldsIndex := FieldsIndex;
    if Active then
     begin
       UpdateCursorPos;
       SelectIndex;
       Resync([])
     end
  end
end;

function TClarionTable.CreateKey(Buf: PChar): string;
var i, j, p: Integer;
    b: Pointer;
begin
  p:=1;
  SetLength(Result, FActiveKey.FKeyRecord.CompLen);
  for i:=0 to FActiveKey.FKeyRecord.NumComps-1 do
   begin
     b:=@Buf[FActiveKey.FKeyItems[i].ElmOff+Sizeof(TDataHeader)];
     case FActiveKey.FKeyItems[i].FldType of
      FLD_GROUP, FLD_STRING, FLD_PICTURE:
       begin
         Move(b^, Result[p], FActiveKey.FKeyItems[i].ElmLen);
         if FActiveKey.FKeyRecord.CompType and KEY_UPRSW<>0 then
          for j:=p to p + FActiveKey.FKeyItems[i].ElmLen do
           Result[j]:=UpCase(Result[j]);
         if FHeader.SFAtr AND SIGN_ENCRYPTED <> 0 then
          for j:=p to p + FActiveKey.FKeyItems[i].ElmLen do dec(Result[j]);
         Inc(p, FActiveKey.FKeyItems[i].ElmLen)
       end;
      FLD_BYTE:
       begin
         Result[p]:=Char(b^);
         Inc(p)
       end;
      FLD_SHORT:
       begin
         PWord(@Result[p])^:=Swap(Word(b^) xor $8000);
         Inc(p, 2)
       end;
      FLD_LONG:
       begin
         PDWord(@Result[p])^:=(
           ((DWord(b^)shr 24)and $000000FF)or
           ((DWord(b^)shr 8)and $0000FF00) or
           ((DWord(b^)shl 8)and $00FF0000) or
           (DWord(b^)shl 24))xor $00000080;
         Inc(p, 4)
       end;
      FLD_DECIMAL:
       begin
         Move(b^, Result[p], FActiveKey.FKeyItems[i].ElmLen);
         if Byte(Result[p]) and $80 =0
          then byte(Result[p]):=byte(Result[p])or $80 else
          for j:=p to p + FActiveKey.FKeyItems[i].ElmLen do
           byte(Result[j]):=not byte(Result[j]);
         Inc(p, FActiveKey.FKeyItems[i].ElmLen)
       end;
      FLD_REAL:
       begin
         if Byte(PChar(b)[7]) and $80 <>0
          then for j:=0 to 7 do Byte(Result[p+j]):=not Byte(PChar(b)[7-j]) else
          begin
            Byte(Result[p]):=Byte(PChar(b)[7]) xor $80;
            for j:=1 to 7 do Byte(Result[p+j]):=not Byte(PChar(b)[7-j])
          end;
         Inc(p, 8)
       end;
     end
   end
end;

procedure TClarionTable.SetKeyFields(const Values: array of const);
var
  I: Integer;
  SaveState: TDataSetState;
begin
  if FActiveKey=nil then DatabaseError(erNoIndex, Self);
  SaveState := SetTempState(dsSetKey);
  try
    ClearRecord(FKeyBuffer);
    if High(FActiveKey.FKeyItems) < High(Values) then
     DatabaseError(SFieldIndexError, Self);
    for I := 0 to High(Values) do
     FieldByNumber(FActiveKey.FKeyItems[i].FldNum).AssignValue(Values[I]);
  finally
    RestoreState(SaveState);
  end;
end;

function TClarionTable.FindKey(const KeyValues: array of const): Boolean;
begin
  CheckBrowseMode;
  SetKeyFields(KeyValues);
  Result := GotoKey
end;

procedure TClarionTable.FindNearest(const KeyValues: array of const);
begin
  CheckBrowseMode;
  SetKeyFields(KeyValues);
  GotoNearest
end;

procedure TClarionTable.SetKeyBuffer(Clear: Boolean);
begin
  CheckBrowseMode;
  Move(FKeyBuffer^, FSaveKeyBuffer^, FHeader.RecLen);
  if Clear then ClearRecord(FKeyBuffer);
  SetState(dsSetKey);
  DataEvent(deDataSetChange, 0)
end;

procedure TClarionTable.SetKey;
begin
  SetKeyBuffer(True)
end;

procedure TClarionTable.EditKey;
begin
  SetKeyBuffer(False)
end;

function TClarionTable.GotoKey: Boolean;
var R: LongInt;
begin
  CheckBrowseMode;
  DoBeforeScroll;
  CursorPosChanged;
  if FActiveKey=nil then DatabaseError(erNoIndex);
  R:=FActiveKey.Key.Locate(PChar(CreateKey(FKeyBuffer)));
  Result := R<>0;
  if Result then
   begin
     FCurRec:=R;
     Resync([rmExact, rmCenter]);
     DoAfterScroll
   end
end;

procedure TClarionTable.GotoNearest;
begin
  CheckBrowseMode;
  CursorPosChanged;
  if FActiveKey=nil then DatabaseError(erNoIndex);
  FActiveKey.Key.Locate(PChar(CreateKey(FKeyBuffer)));
  FCurRec:=FActiveKey.Key.CurRec;
  Resync([rmCenter])
end;

{-----------------------------------------------------}

function TClarionTable.Lookup(const KeyFields: string; const KeyValues: Variant;
  const ResultFields: string): Variant;
begin
  Result := Null;
  if LocateRecord(KeyFields, KeyValues, [], False) then
  begin
    SetTempState(dsCalcFields);
    try
      CalculateFields(TempBuffer);
      Result := FieldValues[ResultFields]
    finally
      RestoreState(dsBrowse)
    end;
  end;
end;

function TClarionTable.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
  DoBeforeScroll;
  Result := LocateRecord(KeyFields, KeyValues, Options, True);
  if Result then
  begin
    Resync([rmExact, rmCenter]);
    DoAfterScroll
  end
end;

function TClarionTable.LocateRecord(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions; SyncCursor: Boolean): Boolean;
begin
  Result:=false;
end;

END.

